# -*- coding: utf-8 -*-
from __future__ import absolute_import
import functools
import warnings


def deprecated(func):
    '''This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.'''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "Call to deprecated function {}.".format(func.__name__),
            category=DeprecationWarning,
            filename=func.func_code.co_filename,
            lineno=func.func_code.co_firstlineno + 1
        )
        return func(*args, **kwargs)
    return new_func


from __future__ import absolute_import, unicode_literals

import os

AUTH_URL = "https://auth.kyper.co/"
AWS_WEB_SERVICE = "http://aws.kyper.co"
KYPER_DB_LOCATION = os.path.join(os.path.expanduser("~"), ".kyper.db")



# -*- coding: utf-8 -*-
from __future__ import (absolute_import, division,
                                print_function, unicode_literals)
from builtins import *
from future.standard_library import install_aliases
install_aliases()

import json
import logging
import os
import requests
import getpass
import sqlite3
from urllib import parse as urlparse
from functools import update_wrapper

from kyper.util import config
from kyper.util.api.errors import InvalidCredentialsException

_DB_CONN = sqlite3.connect(config.KYPER_DB_LOCATION)

class AuthenticationFailed(Exception):
    pass

class InvalidRefreshToken(Exception):
    pass

class AuthDBDoesNotExist(Exception):
    pass


def _initialize_tokens(username, auth_token, refresh_token):
    cur = _DB_CONN.cursor()
    sql = """CREATE TABLE IF NOT EXISTS credentials(id INT PRIMARY KEY, username TEXT, auth_token TEXT, refresh_token TEXT)"""
    cur.execute(sql)
    sql = """INSERT OR REPLACE INTO credentials VALUES (1, ?, ?, ?)"""
    cur.execute(sql, (username, auth_token, refresh_token))
    _DB_CONN.commit()
    cur.close()


def _get_tokens_from_username_and_password(username, password):
    r = requests.post(urlparse.urljoin(config.AUTH_URL, "authtoken/"), data=json.dumps({
        "email": username, 
        "password": password}),
        headers={'Content-type': 'application/json', 'Accept': 'application/json'},
        allow_redirects=False
    )
    js = r.json()
    if js['status'] == 'ok':
        return (js["auth_token"], js["refresh_token"])
    elif js['name'] == 'AuthenticationFailed':
        raise AuthenticationFailed
    else:
        raise Exception("{}: {}".format(js['name'], js['detail']))


def _get_tokens_from_auth_and_refresh_tokens(username, auth_token, refresh_token):
    r = requests.post(urlparse.urljoin(config.AUTH_URL, "authtoken/"), data=json.dumps({
        "email": username, 
        "auth_token": auth_token, 
        "refresh_token": refresh_token}),
        headers={'Content-type': 'application/json', 'Accept': 'application/json'},
        allow_redirects=False
    )
    js = r.json()

    if js['status'] == 'ok':
        return (js["auth_token"], js["refresh_token"])
    elif js['name'] == 'InvalidRefreshToken':
        raise InvalidRefreshToken
    else:
        raise Exception("{}: {}".format(js['name'], js['detail']))


def _clear_tokens():
    cur = _DB_CONN.cursor()
    sql = """DROP TABLE IF EXISTS credentials"""
    cur.execute(sql)
    _DB_CONN.commit()
    cur.close()


def save_tokens(username, auth_token, refresh_token):
    sql = """UPDATE credentials SET username=?, auth_token=?, refresh_token=? WHERE id=1"""
    cur = _DB_CONN.cursor()
    try:
        cur.execute(sql, (username, auth_token, refresh_token))
    except sqlite3.OperationalError:
        _initialize_tokens(username, auth_token, refresh_token)
    _DB_CONN.commit()
    cur.close()

def get_auth_token():
    sql = """SELECT auth_token FROM credentials"""
    cur = _DB_CONN.cursor()
    try:
        row = cur.execute(sql).fetchone()
    except sqlite3.OperationalError:
        raise AuthDBDoesNotExist
    
    if row is None:
        raise AuthDBDoesNotExist
    res = row[0]
    cur.close()
    return res

def get_refresh_token():
    sql = """SELECT refresh_token FROM credentials WHERE id=1"""
    cur = _DB_CONN.cursor()
    try:
        row = cur.execute(sql).fetchone()
    except sqlite3.OperationalError:
        raise AuthDBDoesNotExist

    if row is None:
        raise AuthDBDoesNotExist
    res = row[0]
    cur.close()
    return res

def get_username():
    sql = """SELECT username FROM credentials WHERE id=1"""
    cur = _DB_CONN.cursor()
    try:
        row = cur.execute(sql).fetchone()
    except sqlite3.OperationalError:
        raise AuthDBDoesNotExist

    if row is None:
        raise AuthDBDoesNotExist
    res = row[0]
    cur.close()
    return res


def get_new_credentials():
    '''
        get new authentication/refresh tokens via email/password.
        should only be used if refresh token is expired.
    '''
    print("Please enter your Kyper username and password for access.")
    
    while True:
        username = input("Username: ")
        password = getpass.getpass()
        try:
            auth_token, refresh_token = _get_tokens_from_username_and_password(username, password)
        except AuthenticationFailed:
            print("Your username/password combo was incorrect. Retry:")
            continue

        save_tokens(username, auth_token, refresh_token)
        break

    
def refresh_credentials():
    '''
        perform a global refresh of authentication tokens. requires that a valid refresh token and its associated auth
        token be stored in the in-memory DB
    '''
    username = get_username()
    auth_token, refresh_token = _get_tokens_from_auth_and_refresh_tokens(
        get_username(), 
        get_auth_token(),
        get_refresh_token()
    )

    save_tokens(username, auth_token, refresh_token)


def refresh_stale_credentials(func):
    def catch_and_call(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except InvalidCredentialsException as e:
            logging.warning("Invalid tokens, trying a credentials refresh")
            try:
                refresh_credentials()
            except Exception as e:
                if str(e).split(":")[0] == "InvalidRefreshToken":
                    get_new_credentials()

            return func(*args, **kwargs)

    return update_wrapper(catch_and_call, func)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import json
import logging
from functools import update_wrapper

from parse import parse

from flask import Response

class APIException(Exception):
    message_format = "Undefined"

class MissingParameterException(APIException):
    message_format = "Missing parameter: {}"
    def __init__(self, message):
        APIException.__init__(self, self.message_format.format(message))

class InvalidParameterException(APIException):
    message_format = "Invalid parameter: {}"
    def __init__(self, message):
        APIException.__init__(self, self.message_format.format(message))

class InvalidCredentialsException(APIException):
    message_format = "Invalid authentication credentials"
    def __init__(self, message=None):
        APIException.__init__(self, self.message_format.format(message))

class NotAuthedException(APIException):
    message_format = "User not authorized: {}"
    def __init__(self, message=None):
        APIException.__init__(self, self.message_format.format(message))

class PermissionDeniedException(APIException):
    message_format = "You do not have permission to use {}"
    def __init__(self, message):
        APIException.__init__(self, self.message_format.format(message))

class DataServiceNotFoundException(APIException):
    message_format = "Data service not found: service {}, version {}"
    def __init__(self, service, version):
        APIException.__init__(self, self.message_format.format(service, version))

class UnsupportedVersionException(APIException):
    message_format = "Unsupported version: {}"
    def __init__(self, version):
        APIException.__init__(self, self.message_format.format(version))

class UnsupportedMethodException(APIException):
    message_format = "Unsupported method: {}"
    def __init__(self, method):
        APIException.__init__(self, self.message_format.format(method))

class MalformedResponseException(APIException):
    message_format = "The response from the data service was malformed: {}"
    def __init__(self, response):
        APIException.__init__(self, self.message_format.format(response))

class VersionMismatchException(APIException):
    message_format = "The version code between the client ({}) and the data service ({}) don't match"
    def __init__(self, client_version, service_version):
        APIException.__init__(self, self.message_format.format(client_version, service_version))

class RateLimitExceededException(APIException):
    message_format = "You have exceeded the rate limit"
    def __init__(self):
        APIException.__init__(self, self.message_format)

class QueryTooLargeException(APIException):
    message_format = "Your query asks for too much data. You can only request a maximum of {} per query."
    def __init__(self, maximum):
        APIException.__init__(self, self.message_format.format(maximum))

class TestErrorException(APIException):
    message_format = "This is a test error. Message: {}"
    def __init__(self, message=None):
        APIException.__init__(self, self.message_format.format(message))

class UnknownErrorException(APIException):
    message_format = "{}"
    def __init__(self, message):
        APIException.__init__(self, self.message_format.format(message))

CODE_MAP = {
    'missing_parameter': MissingParameterException,
    'invalid_parameter': InvalidParameterException,
    'invalid_credentials': InvalidCredentialsException,
    'not_authed': NotAuthedException,
    'permission_denied': PermissionDeniedException,
    'data_service_not_found': DataServiceNotFoundException,
    'unsupported_version': UnsupportedVersionException,
    'unsupported_method': UnsupportedMethodException,
    'malformed_response': MalformedResponseException,
    'version_mismatch': VersionMismatchException,
    'rate_limit_exceeded': RateLimitExceededException,
    'query_too_large': QueryTooLargeException,
    'test_error': TestErrorException,
    'unknown_error': UnknownErrorException
}

def get_exception(code):
    return CODE_MAP.get(code)

def get_code(exception):
    exception_class = type(exception)
    if issubclass(exception_class, APIException):
        for key, value in CODE_MAP.items():
            if exception_class == value:
                return key

    return None

def make_exception_from_code_and_message(code, message):
    exception_class = get_exception(code)
    if exception_class is None:
        return UnknownErrorException("code: {}, message: {}".format(code, message))

    args = list(parse(exception_class.message_format, message))

    return exception_class(*args)


def make_error(code, message):
    return Response(json.dumps({"status": "error", "code": code, "message": message}), mimetype="application/json")

def make_error_restful(code, message):
    return {"status": "error", "code": code, "message": message}, 400

### Flask exception wrappers ###
def catch_exceptions(func):
    def catch_and_call(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.exception(func.__name__ + " failed")

            code = get_code(e) or "unknown_error"

            message = type(e).__name__ + ": " + e.__str__() if code == "unknown_error" else e.__str__()

            return make_error(code, message)
    return update_wrapper(catch_and_call, func)

def catch_exceptions_restful(func):
    def catch_and_call(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            logging.exception(func.__name__ + " failed")

            code = get_code(e) or "unknown_error"

            message = type(e).__name__ + ": " + e.__str__() if code == "unknown_error" else e.__str__()

            return make_error_restful(code, message)
    return update_wrapper(catch_and_call, func)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import time
import pytz
from pandas import date_range
from pandas import Series
import numpy as np

def datetime_to_epoch(dt):
    from datetime import datetime

    if dt.tzinfo is None:
        raise Exception("Cannot perform conversion on a timezone-naive datetime")

    d = (dt - UTC(1970, 1, 1))
    return int(d.total_seconds())

def datetime_to_seconds(dt):
    from datetime import datetime

    if dt.tzinfo is None:
        raise Exception("Cannot perform conversion on a timezone-naive datetime")

    d = (dt - UTC(1970, 1, 1))
    return int(d.total_seconds())

def datetime_to_milliseconds(dt):
    from datetime import datetime

    if dt.tzinfo is None:
        raise Exception("Cannot perform conversion on a timezone-naive datetime")

    d = (dt - UTC(1970, 1, 1))
    return int(d.total_seconds() * 1e3)

def datetime_to_microseconds(dt):
    from datetime import datetime

    if dt.tzinfo is None:
        raise Exception("Cannot perform conversion on a timezone-naive datetime")

    d = (dt - UTC(1970, 1, 1))
    return int(d.total_seconds() * 1e6)

def timedelta_to_seconds(td):
    return int(td.total_seconds())

def timedelta_to_milliseconds(td):
    return int(td.total_seconds() * 1e3)

def timedelta_to_microseconds(td):
    return int(td.total_seconds() * 1e6)


def eastern_time(*args):
    from datetime import datetime
    eastern = pytz.timezone('US/Eastern') 
    if len(args) == 1 and isinstance(args[0], datetime):
        if not args[0].tzinfo:
            return eastern.localize(args[0])
        else:
            return args[0].astimezone(eastern)
    return eastern.localize(datetime(*args))

def pacific_time(*args):
    from datetime import datetime
    pacific = pytz.timezone('US/Pacific') 
    if len(args) == 1 and isinstance(args[0], datetime):
        if not args[0].tzinfo:
            return pacific.localize(args[0])
        else:
            return args[0].astimezone(pacific)
    return pacific.localize(datetime(*args))

def central_time(*args):
    from datetime import datetime
    central = pytz.timezone('US/Central') 
    if len(args) == 1 and isinstance(args[0], datetime):
        if not args[0].tzinfo:
            return central.localize(args[0])
        else:
            return args[0].astimezone(central)
    return central.localize(datetime(*args))

def mountain_time(*args):
   from datetime import datetime
   mountain = pytz.timezone('US/Mountain') 
   if len(args) == 1 and isinstance(args[0], datetime):
      if not args[0].tzinfo:
         return mountain.localize(args[0])
      else:
         return args[0].astimezone(mountain)
   return mountain.localize(datetime(*args))

def taipei_time(*args):
    from datetime import datetime
    taipei = pytz.timezone('Asia/Taipei') 
    if len(args) == 1 and isinstance(args[0], datetime):
        if not args[0].tzinfo:
            return taipei.localize(args[0])
        else:
            return args[0].astimezone(taipei)
    return taipei.localize(datetime(*args))

def UTC(*args):
    from datetime import datetime 
    if len(args) == 1 and isinstance(args[0], datetime):
        if not args[0].tzinfo:
            return pytz.utc.localize(args[0])
        else:
            return args[0].astimezone(pytz.utc)
    return pytz.utc.localize(datetime(*args))

def eastern_datetime_to_UTC(datetime):
    edt = datetime
    if not datetime.tzinfo:
        eastern = pytz.timezone('US/Eastern') 
        edt = eastern.localize(datetime)
    return edt.astimezone(pytz.utc)

def taipei_datetime_to_UTC(datetime):
    nst = datetime
    if not datetime.tzinfo:
        taipei = pytz.timezone('Asia/Taipei')
        nst = taipei.localize(datetime)
    return nst.astimezone(pytz.utc)

def UTC_datetime_to_eastern(utc_datetime):
    udt = utc_datetime
    if not utc_datetime.tzinfo:
        udt = pytz.utc.localize(utc_datetime)
    eastern = pytz.timezone('US/Eastern') 
    return udt.astimezone(eastern)

def UTC_datetime_to_taipei(utc_datetime):
    udt = utc_datetime
    if not utc_datetime.tzinfo:
        udt = pytz.utc.localize(utc_datetime)
    taipei = pytz.timezone('Asia/Taipei') 
    return udt.astimezone(taipei)

def eastern_datetime_to_mysql_str(datetime):
    utc_dt = easternDatetimeToUTC(datetime)
    return str(utc_dt).replace('+00:00','')

def freq_to_timedelta(freq):
    if not freq:
        import datetime
        return datetime.timedelta(milliseconds=1)
    d = date_range('1/1/2012', periods=2, freq=freq)
    return d[1] - d[0]

def freq_to_timedelta64(freq):
    freq = freq_to_timedelta(freq)
    return np.array([freq], dtype="timedelta64[ms]")[0]

def make_every(freq):
    """
        This function returns True for every <freq>.

        freq is a datetime.timedelta.
    """
    from datetime import datetime
    def _every():
        now = datetime.now()
        if now - _every.last_call >= freq:
            _every.last_call = now
            return True
        return False
    _every.last_call = datetime.now()
    return _every

def get_gaps_in_timeseries(datetime_idx):
    if np.__version__ >= '1.7':
        diff_dt = Series(np.diff(datetime_idx.values))
    else:
        diff_dt = Series(datetime_idx.to_pydatetime()).diff()
    return diff_dt


def get_next_weekday(start):
    from datetime import timedelta
    start += timedelta(1)
    if start.weekday() > 4:
        start += timedelta(7 - start.weekday())
    return start


def get_prev_weekday(start):
    from datetime import timedelta
    start -= timedelta(1)
    if start.weekday() > 4:
        start -= timedelta(start.weekday() - 4)
    return start


def get_closest_weekday(start, forward=True):
    if start.weekday() <= 4:
        return start
    elif forward is True:
        return get_next_weekday(start)
    else:
        return get_prev_weekday(start)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

ns = {}
with open(os.path.join(here, 'kyper-util', 'version.py')) as f:
   exec(f.read(), {}, ns)

setup(
   name='kyper.util',
   version=ns['__version__'],
   author='Kyper Developers',
   author_email='developers@kyperdata.com',
   packages=find_packages(),
   url='https://git.kyper.co/kyper/kyper-util',
   description='Kyper Utilities',
   install_requires=[
      "kyper",
      "Flask",
      "parse",
      "numpy",
      "pandas",
      "pytz",
      "requests",
      "boto",
      "future==0.14.3"
   ]
)
